package ru.dbotthepony.kstarbound.client.gl.shader

import it.unimi.dsi.fastutil.objects.Object2ObjectFunction
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import org.lwjgl.opengl.GL46.*
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.checkForGLError
import ru.dbotthepony.kstarbound.client.gl.shader.GLShader
import ru.dbotthepony.kstarbound.client.gl.shader.GLUniformLocation
import ru.dbotthepony.kstarbound.client.gl.shader.ShaderLinkException
import ru.dbotthepony.kvector.api.IFloatMatrix
import ru.dbotthepony.kvector.api.IStruct3f
import ru.dbotthepony.kvector.api.IStruct4f
import java.util.*
import java.util.stream.Stream
import kotlin.collections.HashSet

open class GLShaderProgram(val state: GLStateTracker, shaders: Stream<GLShader>) {
	constructor(state: GLStateTracker, shaders: Collection<GLShader>) : this(state, shaders.stream())
	constructor(state: GLStateTracker, vararg shaders: GLShader) : this(state, Arrays.stream(shaders))

	init {
		state.ensureSameThread()
	}

	val pointer = glCreateProgram()

	init {
		checkForGLError()

		for (shader in shaders) {
			glAttachShader(pointer, shader.pointer)
		}

		glLinkProgram(pointer)

		val success = intArrayOf(0)
		glGetProgramiv(pointer, GL_LINK_STATUS, success)

		if (success[0] == 0) {
			throw ShaderLinkException(glGetProgramInfoLog(pointer))
		}

		glGetError()
	}

	private val locationCache = Object2ObjectOpenHashMap<String, Optional<GLUniformLocation>>()

	/**
	 * Возвращает GLUniformLocation или null, если у данной программы нет такого uniform
	 *
	 * В т.ч. если она была в коде шейдера, но нигде не использовалась (отсечена компилятором)
	 *
	 * Результат поиска кешируется, но для повышения производительности вызывающий код желательно так же
	 * должен кешировать результат (как локальное свойство или переменная, в случае многократного использования)
	 */
	operator fun get(name: String): GLUniformLocation? {
		state.ensureSameThread()

		return locationCache.computeIfAbsent(name, Object2ObjectFunction {
			val location = glGetUniformLocation(pointer, name)

			if (location == -1)
				return@Object2ObjectFunction Optional.empty<GLUniformLocation>()

			return@Object2ObjectFunction Optional.of(GLUniformLocation(this, name, location))
		}).orElse(null)
	}

	operator fun set(name: String, value: IStruct4f) = this[name]?.set(value)
	operator fun set(name: String, value: IStruct3f) = this[name]?.set(value)
	operator fun set(name: String, value: Int) = this[name]?.set(value)
	operator fun set(name: String, value: IFloatMatrix<*>) = this[name]?.set(value)

	fun use() = state.use(this)
}
